/*
 * Copyright (C) 2022 - 2026, Stephan Mueller <smueller@chronox.de>
 *
 * License: see LICENSE file in root directory
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, ALL OF
 * WHICH ARE HEREBY DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF NOT ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/*
 * This code is derived in parts from the code distribution provided with
 * https://github.com/kokke/tiny-AES-c
 *
 * This is free and unencumbered software released into the public domain.
 */

#include "aes_aesni.h"
#include "aes_armce.h"
#include "aes_c.h"
#include "aes_riscv64.h"
#include "aes_internal.h"
#include "lc_aes.h"
#include "lc_init.h"
#include "mode_ctr.h"
#include "compare.h"
#include "test_helper_common.h"
#include "timecop.h"
#include "ret_checkers.h"
#include "visibility.h"

#define LC_EXEC_ONE_TEST(aes_impl)                                             \
	if (aes_impl) {                                                        \
		ret += test_encrypt_ctr(aes_impl, #aes_impl);                  \
		ret += test_decrypt_ctr(aes_impl, #aes_impl);                  \
	}

static const uint8_t key_wrap[] = { 0xC9, 0x83, 0xA6, 0xC9, 0xEC, 0x0F, 0x32,
				    0x55, 0x0F, 0x32, 0x55, 0x78, 0x9B, 0xBE,
				    0x78, 0x9B, 0xBE, 0xE1, 0x04, 0x27, 0xE1,
				    0x04, 0x27, 0x4A, 0x6D, 0x90, 0x4A, 0x6D,
				    0x90, 0xB3, 0xD6, 0xF9 };
static const uint8_t in_wrap[] = {
	0x50, 0xB9, 0x22, 0xAE, 0x17, 0x80, 0x0C, 0x75, 0xDE, 0x47, 0xD3, 0x3C,
	0xA5, 0x0E, 0x9A, 0x03, 0x6C, 0xF8, 0x61, 0xCA, 0x33, 0xBF, 0x28, 0x91,
	0x1D, 0x86, 0xEF, 0x58, 0xE4, 0x4D, 0xB6, 0x1F, 0xAB, 0x14, 0x7D, 0x09,
	0x72, 0xDB, 0x44, 0xD0, 0x39, 0xA2, 0x0B, 0x97, 0x00, 0x69, 0xF5, 0x5E,
	0xC7, 0x30, 0xBC, 0x25, 0x8E, 0x1A, 0x83, 0xEC, 0x55, 0xE1, 0x4A, 0xB3,
	0x1C, 0xA8, 0x11, 0x7A, 0x06, 0x6F, 0xD8, 0x41, 0xCD, 0x36, 0x9F, 0x08,
	0x94, 0xFD, 0x66, 0xF2, 0x5B, 0xC4, 0x2D, 0xB9, 0x22, 0x8B, 0x17, 0x80,
	0xE9, 0x52, 0xDE, 0x47, 0xB0, 0x19, 0xA5, 0x0E, 0x77, 0x03, 0x6C, 0xD5,
	0x3E, 0xCA, 0x33, 0x9C, 0x05, 0x91, 0xFA, 0x63, 0xEF, 0x58, 0xC1, 0x2A,
	0xB6, 0x1F, 0x88, 0x14, 0x7D, 0xE6, 0x4F, 0xDB, 0x44, 0xAD, 0x16, 0xA2,
	0x0B, 0x74, 0x00, 0x69, 0xD2, 0x3B, 0xC7, 0x30, 0x99, 0x02, 0x8E, 0xF7,
	0x60, 0xEC, 0x55, 0xBE, 0x27, 0xB3, 0x1C, 0x85, 0x11, 0x7A, 0xE3, 0x4C,
	0xD8, 0x41, 0xAA, 0x13, 0x9F, 0x08, 0x71, 0xFD, 0x66, 0xCF, 0x38, 0xC4,
	0x2D, 0x96, 0x22, 0x8B, 0xF4, 0x5D, 0xE9, 0x52, 0xBB, 0x24, 0xB0, 0x19,
	0x82, 0x0E, 0x77, 0xE0, 0x49, 0xD5, 0x3E, 0xA7, 0x10, 0x9C, 0x05, 0x6E,
	0xFA, 0x63, 0xCC, 0x35, 0xC1, 0x2A, 0x93, 0x1F, 0x88, 0xF1, 0x5A, 0xE6,
	0x4F, 0xB8, 0x21, 0xAD, 0x16, 0x7F, 0x0B, 0x74, 0xDD, 0x46, 0xD2, 0x3B,
	0xA4, 0x0D, 0x99, 0x02, 0x6B, 0xF7, 0x60, 0xC9, 0x32, 0xBE, 0x27, 0x90,
	0x1C, 0x85, 0xEE, 0x57, 0xE3, 0x4C, 0xB5, 0x1E, 0xAA, 0x13, 0x7C, 0x08,
	0x71, 0xDA, 0x43, 0xCF, 0x38, 0xA1, 0x0A, 0x96, 0xFF, 0x68, 0xF4, 0x5D,
	0xC6, 0x2F, 0xBB, 0x24, 0x8D, 0x19, 0x82, 0xEB, 0x54, 0xE0, 0x49, 0xB2,
	0x1B, 0xA7, 0x10, 0x79, 0x05, 0x6E, 0xD7, 0x40, 0xCC, 0x35, 0x9E, 0x07,
	0x93, 0xFC, 0x65, 0xF1, 0x5A, 0xC3, 0x2C, 0xB8, 0x21, 0x8A, 0x16, 0x7F,
	0xE8, 0x51, 0xDD, 0x46, 0xAF, 0x18, 0xA4, 0x0D, 0x76, 0x02, 0x6B, 0xD4,
	0x3D, 0xC9, 0x32, 0x9B, 0x04, 0x90, 0xF9, 0x62, 0xEE, 0x57, 0xC0, 0x29,
	0xB5, 0x1E, 0x87, 0x13, 0x7C, 0xE5, 0x4E, 0xDA, 0x43, 0xAC, 0x15, 0xA1,
	0x0A, 0x73, 0xFF, 0x68, 0xD1, 0x3A, 0xC6, 0x2F, 0x98, 0x01, 0x8D, 0xF6,
	0x5F, 0xEB, 0x54, 0xBD, 0x26, 0xB2, 0x1B, 0x84, 0x10, 0x79, 0xE2, 0x4B,
	0xD7, 0x40, 0xA9, 0x12, 0x9E, 0x07, 0x70, 0xFC, 0x65, 0xCE, 0x37, 0xC3,
	0x2C, 0x95, 0x21, 0x8A, 0xF3, 0x5C, 0xE8, 0x51, 0xBA, 0x23, 0xAF, 0x18,
	0x81, 0x0D, 0x76, 0xDF, 0x48, 0xD4, 0x3D, 0xA6, 0x0F, 0x9B, 0x04, 0x6D,
	0xF9, 0x62, 0xCB, 0x34, 0xC0, 0x29, 0x92, 0x1E, 0x87, 0xF0, 0x59, 0xE5,
	0x4E, 0xB7, 0x20, 0xAC, 0x15, 0x7E, 0x0A, 0x73, 0xDC, 0x45, 0xD1, 0x3A,
	0xA3, 0x0C, 0x98, 0x01, 0x6A, 0xF6, 0x5F, 0xC8, 0x31, 0xBD, 0x26, 0x8F,
	0x1B, 0x84, 0xED, 0x56, 0xE2, 0x4B, 0xB4, 0x1D, 0xA9, 0x12, 0x7B, 0x07,
	0x70, 0xD9, 0x42, 0xCE, 0x37, 0xA0, 0x09, 0x95, 0xFE, 0x67, 0xF3, 0x5C,
	0xC5, 0x2E, 0xBA, 0x23, 0x8C, 0x18, 0x81, 0xEA, 0x53, 0xDF, 0x48, 0xB1,
	0x1A, 0xA6, 0x0F, 0x78, 0x04, 0x6D, 0xD6, 0x3F, 0xCB, 0x34, 0x9D, 0x06,
	0x92, 0xFB, 0x64, 0xF0, 0x59, 0xC2, 0x2B, 0xB7, 0x20, 0x89, 0x15, 0x7E,
	0xE7, 0x50, 0xDC, 0x45, 0xAE, 0x17, 0xA3, 0x0C, 0x75, 0x01, 0x6A, 0xD3,
	0x3C, 0xC8, 0x31, 0x9A, 0x03, 0x8F, 0xF8, 0x61, 0xED, 0x56, 0xBF, 0x28,
	0xB4, 0x1D, 0x86, 0x12,
};
static const uint8_t iv_wrap[] = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
				   0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
				   0xFF, 0xFF, 0xFF, 0xFD };
static const uint8_t ct_wrap[] = {
	0x04, 0xF3, 0xD3, 0x88, 0x17, 0xEF, 0xDC, 0xEF, 0x8B, 0x04, 0xF8, 0x3A,
	0x66, 0x8D, 0x1A, 0x53, 0x57, 0x1F, 0x4B, 0x23, 0xE4, 0xA0, 0xAF, 0xF9,
	0x69, 0x95, 0x35, 0x98, 0x8D, 0x4D, 0x8C, 0xC1, 0xF0, 0xB2, 0x7F, 0x80,
	0xBB, 0x54, 0x28, 0xA2, 0x7A, 0x1B, 0x9F, 0x77, 0xEC, 0x0E, 0x6E, 0xDE,
	0xF0, 0xEC, 0xB8, 0xE4, 0x20, 0x62, 0xEE, 0xDB, 0x5D, 0xF5, 0xDD, 0xE3,
	0x54, 0xFC, 0xDD, 0xEB, 0x6A, 0xEE, 0x65, 0xA1, 0x21, 0xD6, 0xD7, 0x81,
	0x47, 0x61, 0x12, 0x4D, 0xC2, 0x8C, 0xFA, 0x78, 0x1F, 0x28, 0x02, 0x01,
	0xC3, 0xFC, 0x1F, 0xEC, 0x0F, 0x10, 0x4F, 0xB3, 0x12, 0x45, 0xC6, 0x3B,
	0x7E, 0x08, 0xF9, 0x5A, 0xD0, 0x5D, 0x73, 0x2D, 0x58, 0xA4, 0xE5, 0xCB,
	0x1C, 0xB4, 0xCE, 0x74, 0x32, 0x41, 0x1F, 0x31, 0x9C, 0x08, 0xA2, 0x5D,
	0x67, 0xEB, 0x72, 0x1D, 0xF8, 0xE7, 0x70, 0x54, 0x34, 0x4B, 0x31, 0x69,
	0x84, 0x66, 0x96, 0x44, 0x56, 0xCC, 0x1E, 0xD9, 0xE6, 0x13, 0x6A, 0xB9,
	0x2D, 0x0A, 0x05, 0x45, 0x2D, 0x90, 0xCC, 0xDF, 0x16, 0x5C, 0x5F, 0x79,
	0x34, 0x52, 0x54, 0xFE, 0xFE, 0xCD, 0xAD, 0x04, 0x2E, 0xAD, 0x86, 0x06,
	0x1F, 0x37, 0xE8, 0x28, 0xBC, 0xD3, 0x8F, 0x5B, 0x92, 0x66, 0x87, 0x3B,
	0x8A, 0x0A, 0x1A, 0xCC, 0x6E, 0xAB, 0x9F, 0x0B, 0xFA, 0x5C, 0xE6, 0xFD,
	0x3C, 0x98, 0x08, 0x12, 0xEC, 0xAA, 0x9E, 0x11, 0xCA, 0xB2, 0x1F, 0xCE,
	0x5E, 0x5B, 0xB2, 0x72, 0x9C, 0xCC, 0x5D, 0xC5, 0xE0, 0x32, 0xC0, 0x56,
	0xD5, 0x45, 0x16, 0xD2, 0xAF, 0x13, 0x66, 0xF7, 0x8C, 0x67, 0xAC, 0x79,
	0xB2, 0xAF, 0x56, 0x27, 0x3F, 0xCC, 0xFE, 0xCB, 0x1E, 0xC0, 0x75, 0xF1,
	0xA7, 0xC9, 0xC3, 0x1D, 0x8E, 0xDD, 0xF9, 0xD4, 0x42, 0xC8, 0x21, 0x08,
	0x16, 0xF7, 0x01, 0xD7, 0xAC, 0x8E, 0x3F, 0x1D, 0x56, 0xC1, 0x06, 0xE4,
	0x9C, 0x62, 0xD6, 0xA5, 0x6A, 0x50, 0x44, 0xB3, 0x35, 0x1C, 0x82, 0xB9,
	0x10, 0xF9, 0x42, 0xA1, 0xFC, 0x74, 0x9B, 0x44, 0x4F, 0x25, 0x02, 0xE3,
	0x08, 0xF5, 0xD4, 0x32, 0x39, 0x08, 0x11, 0xE8, 0xD2, 0x6B, 0x50, 0x53,
	0xD4, 0x08, 0xD1, 0x6B, 0x3A, 0x4A, 0x68, 0x7B, 0x7C, 0xCD, 0x46, 0x5E,
	0x0D, 0x07, 0x19, 0xDB, 0x67, 0xD7, 0x98, 0x91, 0xD7, 0x17, 0x10, 0x9B,
	0x7B, 0x8A, 0x9B, 0x33, 0xAE, 0xF3, 0x00, 0xA6, 0xD4, 0x15, 0xD9, 0xEA,
	0x85, 0x99, 0x22, 0xE8, 0x91, 0x38, 0x70, 0x83, 0x93, 0x01, 0x24, 0x6C,
	0xFA, 0x9A, 0xB9, 0x07, 0xEA, 0x8D, 0x3B, 0xD9, 0x2A, 0x43, 0x59, 0x16,
	0x2F, 0x69, 0xEE, 0x84, 0x36, 0x44, 0x76, 0x98, 0xF3, 0x04, 0x2A, 0x7C,
	0x74, 0x3D, 0x29, 0x2B, 0x0D, 0xAD, 0x8F, 0x44, 0x82, 0x9E, 0x57, 0x8D,
	0xAC, 0xED, 0x18, 0x1F, 0x50, 0xA4, 0xF5, 0x98, 0x1F, 0xBD, 0x92, 0x91,
	0x1B, 0x2D, 0xA6, 0xD6, 0xD2, 0xE3, 0x02, 0xAA, 0x92, 0x3B, 0xC6, 0xB3,
	0x1B, 0x39, 0x72, 0xD5, 0x26, 0xCA, 0x04, 0xE0, 0xFC, 0x58, 0x78, 0xBB,
	0xB1, 0x3F, 0xA1, 0x9C, 0x42, 0x24, 0x3E, 0x2E, 0x22, 0xBB, 0x4B, 0xBA,
	0xF4, 0x52, 0x0A, 0xE6, 0xAE, 0x47, 0xB4, 0x7D, 0x1D, 0xA8, 0xBE, 0x81,
	0x1A, 0x75, 0xDA, 0xAC, 0xA6, 0x25, 0x1E, 0xEF, 0x3A, 0xC0, 0x6C, 0x63,
	0xEF, 0xDC, 0xC9, 0x79, 0x10, 0x26, 0xE8, 0x61, 0x29, 0xFC, 0xA4, 0x05,
	0xDF, 0x7D, 0x5C, 0x63, 0x10, 0x09, 0x9B, 0x46, 0x9B, 0xF2, 0x2C, 0x2B,
	0xFA, 0x3A, 0x05, 0x4C, 0xFA, 0xD1, 0xFF, 0xFE, 0xF1, 0x4C, 0xE5, 0xB2,
	0x91, 0x64, 0x0C, 0x51
};

static const uint8_t key256[] = { 0x60, 0x3d, 0xeb, 0x10, 0x15, 0xca, 0x71,
				  0xbe, 0x2b, 0x73, 0xae, 0xf0, 0x85, 0x7d,
				  0x77, 0x81, 0x1f, 0x35, 0x2c, 0x07, 0x3b,
				  0x61, 0x08, 0xd7, 0x2d, 0x98, 0x10, 0xa3,
				  0x09, 0x14, 0xdf, 0xf4 };
static const uint8_t in256[] = {
	0x60, 0x1e, 0xc3, 0x13, 0x77, 0x57, 0x89, 0xa5, 0xb7, 0xa7, 0xf5,
	0x04, 0xbb, 0xf3, 0xd2, 0x28, 0xf4, 0x43, 0xe3, 0xca, 0x4d, 0x62,
	0xb5, 0x9a, 0xca, 0x84, 0xe9, 0x90, 0xca, 0xca, 0xf5, 0xc5, 0x2b,
	0x09, 0x30, 0xda, 0xa2, 0x3d, 0xe9, 0x4c, 0xe8, 0x70, 0x17, 0xba,
	0x2d, 0x84, 0x98, 0x8d, 0xdf, 0xc9, 0xc5, 0x8d, 0xb6, 0x7a, 0xad,
	0xa6, 0x13, 0xc2, 0xdd, 0x08, 0x45, 0x79, 0x41, 0xa6
};
static const uint8_t key192[] = { 0x8e, 0x73, 0xb0, 0xf7, 0xda, 0x0e,
				  0x64, 0x52, 0xc8, 0x10, 0xf3, 0x2b,
				  0x80, 0x90, 0x79, 0xe5, 0x62, 0xf8,
				  0xea, 0xd2, 0x52, 0x2c, 0x6b, 0x7b };
static const uint8_t in192[] = {
	0x1a, 0xbc, 0x93, 0x24, 0x17, 0x52, 0x1c, 0xa2, 0x4f, 0x2b, 0x04,
	0x59, 0xfe, 0x7e, 0x6e, 0x0b, 0x09, 0x03, 0x39, 0xec, 0x0a, 0xa6,
	0xfa, 0xef, 0xd5, 0xcc, 0xc2, 0xc6, 0xf4, 0xce, 0x8e, 0x94, 0x1e,
	0x36, 0xb2, 0x6b, 0xd1, 0xeb, 0xc6, 0x70, 0xd1, 0xbd, 0x1d, 0x66,
	0x56, 0x20, 0xab, 0xf7, 0x4f, 0x78, 0xa7, 0xf6, 0xd2, 0x98, 0x09,
	0x58, 0x5a, 0x97, 0xda, 0xec, 0x58, 0xc6, 0xb0, 0x50
};
static const uint8_t key128[] = { 0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae,
				  0xd2, 0xa6, 0xab, 0xf7, 0x15, 0x88,
				  0x09, 0xcf, 0x4f, 0x3c };
static const uint8_t in128[] = {
	0x87, 0x4d, 0x61, 0x91, 0xb6, 0x20, 0xe3, 0x26, 0x1b, 0xef, 0x68,
	0x64, 0x99, 0x0d, 0xb6, 0xce, 0x98, 0x06, 0xf6, 0x6b, 0x79, 0x70,
	0xfd, 0xff, 0x86, 0x17, 0x18, 0x7b, 0xb9, 0xff, 0xfd, 0xff, 0x5a,
	0xe4, 0xdf, 0x3e, 0xdb, 0xd5, 0xd3, 0x5e, 0x5b, 0x4f, 0x09, 0x02,
	0x0d, 0xb0, 0x3e, 0xab, 0x1e, 0x03, 0x1d, 0xda, 0x2f, 0xbe, 0x03,
	0xd1, 0x79, 0x21, 0x70, 0xa0, 0xf3, 0x00, 0x9c, 0xee
};

static const uint8_t iv_sizes[] = { 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5,
				    0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb,
				    0xfc, 0xfd, 0xfe, 0xff };

static const uint8_t out_sizes[] = {
	0x6b, 0xc1, 0xbe, 0xe2, 0x2e, 0x40, 0x9f, 0x96, 0xe9, 0x3d, 0x7e,
	0x11, 0x73, 0x93, 0x17, 0x2a, 0xae, 0x2d, 0x8a, 0x57, 0x1e, 0x03,
	0xac, 0x9c, 0x9e, 0xb7, 0x6f, 0xac, 0x45, 0xaf, 0x8e, 0x51, 0x30,
	0xc8, 0x1c, 0x46, 0xa3, 0x5c, 0xe4, 0x11, 0xe5, 0xfb, 0xc1, 0x19,
	0x1a, 0x0a, 0x52, 0xef, 0xf6, 0x9f, 0x24, 0x45, 0xdf, 0x4f, 0x9b,
	0x17, 0xad, 0x2b, 0x41, 0x7b, 0xe6, 0x6c, 0x37, 0x10
};

static int test_xcrypt_ctr_one(const char *xcrypt, struct lc_sym_ctx *ctx,
			       const uint8_t *key, size_t keylen,
			       const uint8_t *iv, size_t ivlen,
			       const uint8_t *in, const uint8_t *out,
			       size_t datalen)
{
	uint8_t act[500];
	char status[64];
	int ret;

	if (sizeof(act) < datalen)
		return -EFAULT;

	snprintf(status, sizeof(status), "AES-CTR %s", xcrypt);

	/* Unpoison key to let implementation poison it */
	unpoison(key, keylen);

	/* Encrypt */
	CKINT(lc_sym_init(ctx));
	CKINT(lc_sym_setkey(ctx, key, keylen));
	CKINT(lc_sym_setiv(ctx, iv, ivlen));
	lc_sym_encrypt(ctx, in, act, datalen);
	ret = lc_compare(act, out, datalen, status);

out:
	return ret;
}

static int test_xcrypt_ctr(const struct lc_sym *aes, const char *name,
			   const char *xcrypt)
{
	struct lc_sym_ctx *aes_ctr_heap;
	int ret;
	LC_SYM_CTX_ON_STACK(aes_ctr, aes);

	printf("AES CTR ctx %s (%s implementation) len %u\n", name,
	       aes == lc_aes_ctr_c ? "C" : "accelerated",
	       (unsigned int)LC_SYM_CTX_SIZE(aes));

	ret = test_xcrypt_ctr_one(xcrypt, aes_ctr, key256, sizeof(key256),
				  iv_sizes, sizeof(iv_sizes), in256, out_sizes,
				  sizeof(in256));
	lc_sym_zero(aes_ctr);

	ret += test_xcrypt_ctr_one(xcrypt, aes_ctr, key192, sizeof(key192),
				   iv_sizes, sizeof(iv_sizes), in192, out_sizes,
				   sizeof(in192));
	lc_sym_zero(aes_ctr);

	ret += test_xcrypt_ctr_one(xcrypt, aes_ctr, key128, sizeof(key128),
				   iv_sizes, sizeof(iv_sizes), in128, out_sizes,
				   sizeof(in128));
	lc_sym_zero(aes_ctr);

	if (lc_sym_alloc(aes, &aes_ctr_heap))
		return ret + 1;
	ret += test_xcrypt_ctr_one(xcrypt, aes_ctr_heap, key256, sizeof(key256),
				   iv_sizes, sizeof(iv_sizes), in256, out_sizes,
				   sizeof(in256));
	lc_sym_zero_free(aes_ctr_heap);

	ret += test_xcrypt_ctr_one(xcrypt, aes_ctr, key_wrap, sizeof(key_wrap),
				   iv_wrap, sizeof(iv_wrap), in_wrap, ct_wrap,
				   sizeof(in_wrap));
	lc_sym_zero(aes_ctr);

	return ret;
}

static int test_encrypt_ctr(const struct lc_sym *aes, const char *name)
{
	return test_xcrypt_ctr(aes, name, "encrypt");
}

static int test_decrypt_ctr(const struct lc_sym *aes, const char *name)
{
	return test_xcrypt_ctr(aes, name, "decrypt");
}

static void ctr_inc(uint8_t *iv)
{
	int bi;

	/* Increment Iv and handle overflow */
	for (bi = (AES_BLOCKLEN - 1); bi >= 0; --bi) {
		/* inc will overflow */
		if (iv[bi] == 255) {
			iv[bi] = 0;
			continue;
		}
		iv[bi] += 1;
		break;
	}
}

#include "../src/asm/AESNI_x86_64/aes_aesni_x86_64.h"
struct lc_sym_state_aesni {
	struct aes_aesni_block_ctx enc_block_ctx;
	uint8_t iv[AES_BLOCKLEN];
};

#include "../src/asm/ARMv8/aes_armv8_ce.h"
struct lc_sym_state_armce {
	struct aes_aesni_block_ctx enc_block_ctx;
	uint8_t iv[AES_BLOCKLEN];
};

static int ctr_tester_one(uint8_t *iv, uint64_t *iv128)
{
	uint8_t buffer64[AES_BLOCKLEN] = { 0 };
	uint8_t data[AES_BLOCKLEN] = { 0 };
	uint8_t key[2 * AES_BLOCKLEN] = { 0 };
	unsigned int i;
	int ret = 0;
	LC_SYM_CTX_ON_STACK(aesni, lc_aes_ctr_aesni);
	LC_SYM_CTX_ON_STACK(aes_armce, lc_aes_ctr_armce);

	CKINT(lc_sym_init(aesni));
	CKINT(lc_sym_setkey(aesni, key, sizeof(key)));
	CKINT(lc_sym_setiv(aesni, iv, AES_BLOCKLEN));

	/* Unpoison key to let implementation poison it */
	unpoison(key, sizeof(key));

	CKINT(lc_sym_init(aes_armce));
	CKINT(lc_sym_setkey(aes_armce, key, sizeof(key)));
	CKINT(lc_sym_setiv(aes_armce, iv, AES_BLOCKLEN));

	for (i = 0; i < 10; i++) {
		ctr_inc(iv);
		ctr128_inc(iv128);
		ctr128_to_ptr(buffer64, iv128);
		CKINT(lc_compare(buffer64, iv, AES_BLOCKLEN,
				 "CTR 64 maintenance"));

		/* Test counter management for AESNI implementation */
		if (lc_aes_ctr_aesni != lc_aes_ctr_c) {
			lc_sym_encrypt(aesni, data, data, sizeof(data));
			CKINT(lc_compare(
				buffer64,
				((struct lc_sym_state_aesni *)(aesni->sym_state))
					->iv,
				AES_BLOCKLEN, "CTR AESNI maintenance"));
		}

		/* Test counter management for ARM-CE implementation */
		if (lc_aes_ctr_armce != lc_aes_ctr_c) {
			lc_sym_encrypt(aes_armce, data, data, sizeof(data));
			CKINT(lc_compare(buffer64,
					 ((struct lc_sym_state_armce
						   *)(aes_armce->sym_state))
						 ->iv,
					 AES_BLOCKLEN,
					 "CTR ARM-CE maintenance"));
		}
	}

out:
	return ret;
}

static int ctr_tester(void)
{
	unsigned int i;
	uint8_t iv[AES_BLOCKLEN];
	uint64_t iv128[AES_CTR128_64BIT_WORDS];
	int ret;

	memset(iv, 0, sizeof(iv));
	memset(iv128, 0, sizeof(iv128));
	ret = ctr_tester_one(iv, iv128);

	memset(iv, 0, sizeof(iv));
	memset(iv128, 0, sizeof(iv128));
	for (i = AES_BLOCKLEN - 1; i >= AES_BLOCKLEN - sizeof(uint64_t); i--)
		iv[i] = 0xff;
	iv128[AES_CTR128_64BIT_WORDS - 1] = (uint64_t)-1;
	ret += ctr_tester_one(iv, iv128);

	memset(iv, 0, sizeof(iv));
	memset(iv128, 0, sizeof(iv128));
	for (i = AES_BLOCKLEN; i > 0; i--)
		iv[i - 1] = 0xff;
	for (i = AES_CTR128_64BIT_WORDS; i > 0; i--)
		iv128[i - 1] = (uint64_t)-1;
	ret += ctr_tester_one(iv, iv128);

	return ret;
}

static int test_xcrypt_ctr_common(void)
{
	uint8_t act[sizeof(in_wrap)];
	int ret;
	LC_AES_CTR_CTX_ON_STACK(aes_ctr);

	/* Unpoison key to let implementation poison it */
	unpoison(key_wrap, sizeof(key_wrap));

	/* Encrypt */
	CKINT(lc_sym_init(aes_ctr));
	CKINT(lc_sym_setkey(aes_ctr, key_wrap, sizeof(key_wrap)));
	CKINT(lc_sym_setiv(aes_ctr, iv_wrap, sizeof(iv_wrap)));
	lc_sym_encrypt(aes_ctr, in_wrap, act, sizeof(in_wrap));
	ret = lc_compare(act, ct_wrap, sizeof(in_wrap), "Common AES-CTR");

out:
	return ret;
}

static int test_ctr(void)
{
	int ret = 0;

	ret += ctr_tester();

	LC_EXEC_ONE_TEST(lc_aes_ctr);
	LC_EXEC_ONE_TEST(lc_aes_ctr_aesni);
	LC_EXEC_ONE_TEST(lc_aes_ctr_armce);
	LC_EXEC_ONE_TEST(lc_aes_ctr_c);
	LC_EXEC_ONE_TEST(lc_aes_ctr_riscv64);
	ret += test_xcrypt_ctr_common();

	return !!ret;
}

LC_TEST_FUNC(int, main, int argc, char *argv[])
{
	int ret;

	(void)argv;

	if (argc >= 2)
		lc_init(LC_INIT_AES_SBOX);

	ret = test_ctr();

	ret = test_validate_status(ret, lc_sym_alg_status(lc_aes_ctr), 1);
	ret = test_validate_status(ret, lc_sym_alg_status(lc_aes_ctr_aesni),
				   lc_aes_ctr == lc_aes_ctr_aesni);
	ret = test_validate_status(ret, lc_sym_alg_status(lc_aes_ctr_armce),
				   lc_aes_ctr == lc_aes_ctr_armce);
	ret = test_validate_status(ret, lc_sym_alg_status(lc_aes_ctr_c),
				   lc_aes_ctr == lc_aes_ctr_c);
	ret = test_validate_status(ret, lc_sym_alg_status(lc_aes_ctr_riscv64),
				   lc_aes_ctr == lc_aes_ctr_riscv64);
	ret += test_print_status();

	return ret;
}
